'''
Created on Feb 26, 2016

@author: alevin
'''
from __future__ import absolute_import

import unittest

import mock.mock as mock
import requests

import pythesint as pti
from pythesint.gcmd_vocabulary import GCMDVocabulary


class GCMDVocabularyTest(unittest.TestCase):

    def test_read_revision(self):
        gcmd_list = []
        line = ('"Keyword Version: 8.1","Revision: 2016-01-08 13:40:40","'
                'Timestamp: 2016-02-26 04:46:02","Terms Of Use: See '
                'http://gcmd.nasa.gov/r/l/TermsOfUse","The most up to date XML'
                ' representations can be found here: '
                'http://gcmd.earthdata.nasa.gov/kms/concepts/concept_scheme'
                '/instruments/?format=xml"')
        pti.gcmd_vocabulary.GCMDVocabulary._read_revision(line, gcmd_list)
        self.assertEqual(len(gcmd_list), 1)
        self.assertDictEqual(
            gcmd_list[0],
            {'Keyword Version': '8.1', 'Revision': '2016-01-08 13:40:40'})

    def test_get_location_by_type(self):
        type = 'africa'
        a = pti.get_gcmd_location(type)
        self.assertEqual(a['Location_Type'], 'AFRICA')

    def test_exception_on_unavailable_remote_file(self):
        voc = GCMDVocabulary(name='test_voc', url='https://sdfghdfghd.nersc.no')
        with self.assertRaises(requests.RequestException):
            voc._fetch_online_data()

    def test_fetch_version(self):
        """Test that the specified version is fetched"""
        voc = GCMDVocabulary(name='test_voc', url='https://sdfghdfghd.nersc.no')
        # We use the exception as side effect to skip the part of the
        # method we are not testing.
        # TODO: split _fetch_online_data() into several methods instead
        with mock.patch('requests.get', side_effect=requests.RequestException) as mock_get:
            with self.assertRaises(requests.RequestException):
                voc._fetch_online_data()
            mock_get.assert_called_with('https://sdfghdfghd.nersc.no', verify=False, params={})

            with self.assertRaises(requests.RequestException):
                voc._fetch_online_data(version=None)
            mock_get.assert_called_with('https://sdfghdfghd.nersc.no', verify=False, params={})

            with self.assertRaises(requests.RequestException):
                voc._fetch_online_data(version='')
            mock_get.assert_called_with('https://sdfghdfghd.nersc.no', verify=False, params={})

            with self.assertRaises(requests.RequestException):
                voc._fetch_online_data(version='9.1.5')
            mock_get.assert_called_with(
                'https://sdfghdfghd.nersc.no', verify=False, params={'version': '9.1.5'})

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



from collections import OrderedDict

from rapidfuzz.fuzz import token_set_ratio
from rapidfuzz.process import extract
from rapidfuzz.utils import default_process


class Vocabulary(object):
    def __init__(self, name, **kwargs):
        self.name = name
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def find_keyword(self, item):
        ''' Return the dictionary containing the given item in the provided list.
        The function returns only the lowest level match, i.e., the
        dictionary where the subgroups of the matching group are empty or the only
        match, if there is just one.

        Examples:

        1) Searching "Active Remote Sensing" in the instruments list will return
            {
                'Category': 'Earth Remote Sensing Instruments',
                'Class': 'Active Remote Sensing',
                'Type': '',
                'Subtype': '',
                'Short_Name': '',
                'Long_Name': ''
            }

        2) Searching "ASAR" in the instruments list will return
            {
                'Category': 'Earth Remote Sensing Instruments',
                'Class': 'Active Remote Sensing',
                'Type': 'Imaging Radars',
                'Subtype': '',
                'Short_Name': 'ASAR',
                'Long_Name': 'Advanced Synthetic Aperature Radar'
            }

        3) Searching "surface_backwards_scattering_coefficient_of_radar_wave" will
        return

        DOCTESTS:

        '''
        best_match = None
        best_match_empty_cells = 0
        for d in self.get_list():
            match = None
            match_empty_cells = 0
            for _, key in enumerate(d.keys()):
                if d[key].upper() == item.upper():
                    match = d
                # sum up number of empty cells after matched cell
                if match == d and len(d[key]) == 0:
                    match_empty_cells += 1
            # if match is found in this row, and
            #     if the best match is not found yet, or
            #     if the best match has less empty cells after matched cell
            # i.e., if we search for the word 'animal', then the dictionary
            # {'animal', ''} is a better match than {'animal', 'cat'}
            if (match is not None and
                (best_match is None or
                 match_empty_cells > best_match_empty_cells)):
                best_match = match
                best_match_empty_cells = match_empty_cells

        if best_match is None:
            raise IndexError('%s is not found in %s!' % (item, self.name))

        return best_match

    def search(self, keyword):
        ''' Search for keyword in a list and return all matches '''
        matches = []
        for d in self.get_list():
            for key in d.keys():
                if keyword.upper() in d[key].upper() and not d in matches:
                    matches.append(d)

        return matches

    def update(self, version=None):
        pass

    def sort_list(self, list):
        retlist = []
        for dd in list:
            line_kw = OrderedDict()
            has_aliases = isinstance(self.categories, dict)
            try:
                for key in self.categories:
                    if has_aliases:
                        # self.categories is a dict defining aliases
                        alias = self.categories.get(key)
                        # if an empty alias is defined, use the original category name
                        field_name = alias if alias else key
                    else:
                        # no alias is defined, just use the category name
                        field_name = key
                    line_kw[field_name] = dd[key]
            except KeyError:
                continue
            retlist.append(line_kw)
        return retlist

    def get_list(self):
        raise NotImplementedError

    def _fuzzy_search(self, search_string, scorer=token_set_ratio, processor=default_process,
                      results_limit=10, min_score=50.0):
        """Perform a fuzzy search on the vocabulary.
        Fully parameterized, meant to be called by self.fuzzy_search()
        """
        terms_list = self.get_list()
        choices = (' '.join(ordered_dict.values()).lower() for ordered_dict in terms_list)
        # returns a list of tuples (choice, similarity, index)
        # similarity is a float in [0.0, 100.0], 100.0 meaning the
        # search string is a subset of the choice string
        results = extract(search_string.lower(), choices,
                          scorer=scorer, processor=processor,
                          limit=results_limit, score_cutoff=min_score)

        return [terms_list[result[2]] for result in results]

    def fuzzy_search(self, search_string):
        """Perform a fuzzy search on the vocabulary terms.
        Uses default parameters, can be overriden in subclasses.
        The default scorer uses token set ratio, which gives the
        highest score when one of the strings is a subset of the other.
        Words order does not matter.
        """
        return self._fuzzy_search(search_string)



from __future__ import absolute_import

import csv
import requests
import warnings
from collections import OrderedDict

from pythesint.json_vocabulary import JSONVocabulary


class GCMDVocabulary(JSONVocabulary):

    def _check_categories(self, categories):
        '''Print a warning if the categories are not the expected ones
        '''
        if set(self.categories) != set(categories):
            mismatch_categories = set(self.categories).difference(set(categories))
            warnings.warn(f'Unknown categories {mismatch_categories} in {self.name}')

    def _fetch_online_data(self, version=None):
        ''' Return list of GCMD standard keywords
            self.url must be set
        '''
        if version:
            params = {'version': version}
        else:
            params = {}

        try:
            r = requests.get(self.url, verify=False, params=params)
            r.raise_for_status()
        except requests.RequestException:
            print("Could not get the vocabulary file at '{}'".format(self.url))
            raise

        lines = r.text.splitlines()
        keywords = []
        # Add version+revision information
        self._read_revision(lines[0], keywords)
        # parse actual CSV contents
        reader = csv.DictReader(lines[1:], dialect='unix', restval='')
        self._check_categories(reader.fieldnames)
        keywords.extend(list(reader))
        # remove UUID and extra fields
        for kw in keywords[1:]:
            for key in ('UUID', None):
                try:
                    del kw[key]
                except KeyError:
                    pass

        return keywords

    @staticmethod
    def _read_revision(line, gcmd_list):
        ''' Reads the line, extracts the Revision into a new dictionary and appends
        it to gcmd_list
        '''
        # TODO: Cast exception if not found?
        if 'Keyword Version' and 'Revision' in line:
            meta = line.split('","')
            gcmd_list.append({
                'Revision': meta[1][10:],
                'Keyword Version': meta[0].split(': ')[1]
            })
